/* -*- mode:c -*-
 *
 * Flex module clamplex.lex
 *
 * Copyright (c) 2002 Raoul M. Gough
 * Copyright (C) 2006 Joel Yliluoma (http://iki.fi/bisqwit/)
 *
 * This material is provided "as is", with absolutely no warranty expressed
 * or implied. Any use is at your own risk.
 *
 * Permission to use or copy this software for any purpose is hereby granted 
 * without fee, provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 * History
 * =======
 * 2002/ 2/ 6	rmg	File creation
 * 2006-10-23 bisqwit   Rewrote ECHO to avoid #include files
 *
 * version: $Id: $
 */

%option noyywrap
%option never-interactive
%option 8bit
/*%option debug*/
%option nomain
%option yylineno

%{

#include <sstream>

#include <stdio.h>
#include "clamp_support.hh"

  using namespace clamp;

#include "clamp.tab.h"

  int resume_initial = 0;

  int openbraces = 0;

  extern "C" { int yylex (); }

  extern bool parserBraceMatch (int);
  
  extern std::stringstream programout;
  extern std::stringstream* yyoutput;
  extern std::string infn, outfn;
  
  static void Echo()
  {
    static bool line_known = true;
    std::string blob(yytext, yytext+yyleng);
    
    if(yyoutput == &programout)
    {
      *yyoutput << blob;
      line_known = false;
    }
    else
    {
      if(!line_known)
      {
        unsigned line_number = yylineno;
        
    #if 0
        *yyoutput << "\n#line " << line_number << " \"" << infn << "\"\n";
    #endif
        line_known = true;
      }
      *yyoutput << blob;
    }
  }

  #define ECHO Echo()
  
%}

%x stringlit
%x charlit
%x comment
%x lambda
%x lambda_def
%x context
%x cvqual_lambda

FSLASH \/
BSLASH \\
DQUOTE \"
SQUOTE \'
STAR   \*
LBRACE \{
RBRACE \}
COLON  \:

%%
      if (resume_initial)
        {
          BEGIN (INITIAL);
          resume_initial = 0;
        }

<INITIAL>{
  {FSLASH}{FSLASH}.*\r?\n ECHO;

  {DQUOTE} {
    ECHO;
    BEGIN(stringlit);
  }

  {SQUOTE} {
    ECHO;
    BEGIN(charlit);
  }

  {FSLASH}{STAR} {
    ECHO;
    BEGIN(comment);
  }
  
  new[ \t\r\n][ \t\r\n]*lambda {
    BEGIN(lambda);
    return TOK_NEW_LAMBDA;
  }

  lambda {
    BEGIN(lambda);
    return TOK_LAMBDA;
  }
  
  const[ \t\r\n][ \t\r\n]*lambda |
  volatile[ \t\r\n][ \t\r\n]*lambda {
    BEGIN(cvqual_lambda);
    /* cut from space */
    yytext[ *yytext=='c' ? 5 : 8 ] = 0;
    return TOK_CV_QUAL;
  }

  __ctx[ \t\r\n]* {
    BEGIN (context);
    return TOK_CONTEXT;
  }

  __ref[ \t\r\n]* {
    BEGIN (context);
    return TOK_CONTEXT_REF;
  }

  {LBRACE} {
    ECHO;
    ++openbraces;
  }

  {RBRACE} {
    if (parserBraceMatch (openbraces--))
      {
	return '}';
      }

    else
      {
	ECHO;
      }
  }
}

<cvqual_lambda>{
  [ \t\r\n]*  { BEGIN(lambda); return TOK_LAMBDA; }
}

<stringlit>{
  [^\"\\]* ECHO;
  {BSLASH}. ECHO;
  {DQUOTE} {
    ECHO;
    BEGIN(INITIAL);
  }
}

<charlit>{
  [^\'\\]* ECHO;
  {BSLASH}. ECHO;
  {SQUOTE} {
    ECHO;
    BEGIN(INITIAL);
  }
}

<comment>{
  {STAR}{FSLASH} {
    ECHO;
    BEGIN(INITIAL);
  }
}

<lambda>{
  const     |
  volatile  return TOK_CV_QUAL;

  char      |
  int       |
  short     |
  long      |
  signed    |
  unsigned  |
  float     |
  double    return TOK_TYPE_KEYWORD;

  typename  return TOK_TYPENAME;

  lambda    return TOK_LAMBDA;

  {COLON}{COLON} return TOK_SCOPE;

  [^()*&,{}= \t\r\n]* return TOK_GENERIC;

  [ \t\r\n]* /* ignore whitespace */

  {LBRACE} {
    ++openbraces;
    return yytext[yyleng-1];
  }

  . return yytext[yyleng-1];
}

<context>{
  \( |
  \) return yytext[yyleng-1];

  [^()]* return TOK_GENERIC;
}

%%

void lexer_reprocess_last ()
{
  yyless(0);
}
